/*
 *  Copyright (C) 2007 - 2012 GeoSolutions S.A.S.
 *  http://www.geo-solutions.it
 * 
 *  GPLv3 + Classpath exception
 * 
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package it.geosolutions.csv2geofence.impl;

import it.geosolutions.csv2geofence.config.model.RuleFileConfig;
import it.geosolutions.csv2geofence.config.model.RuleFileConfig.ServiceRequest;
import it.geosolutions.csv2geofence.config.model.RuleFileConfig.ServiceRequest.Type;
import it.geosolutions.csv2geofence.config.model.internal.RuleOp;
import it.geosolutions.geofence.core.model.enums.GrantType;
import it.geosolutions.geofence.services.rest.model.RESTBatchOperation;
import it.geosolutions.geofence.services.rest.model.RESTInputRule;
import it.geosolutions.geofence.services.rest.model.util.RESTBatchOperationFactory;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;


/**
 * Transforms RuleOps into RESTBatchoperations
 *
 * @author ETj (etj at geo-solutions.it)
 */
public class RulesProcessor {

    private final static Logger LOGGER = LogManager.getLogger(RulesProcessor.class);

    /**
     * @param ops
     * @param availableGroups may be augmented
     * @param ruleMapping
     * @return
     */
    public List<RESTBatchOperation> buildBatchOps(List<RuleOp> ops, Map<String, String> availableGroups, RuleFileConfig cfg) {
        List<RESTBatchOperation> ret = new ArrayList<RESTBatchOperation>(ops.size());

        for (RuleOp op : ops) {
            LOGGER.debug("Preparing for output " + op);
            List<RESTBatchOperation> restOps = buildBatchOperation(op, availableGroups, cfg);
            ret.addAll(restOps);
        }

        return ret;
    }

    /**
     * @param ruleOp
     * @param availableGroups may be augmented
     * @param ruleMapping
     * @return
     */
    protected List<RESTBatchOperation> buildBatchOperation(RuleOp ruleOp, Map<String, String> availableGroups, RuleFileConfig ruleFileConfig) {

        Map<String,List<RuleFileConfig.ServiceRequest>> ruleMapping = ruleFileConfig.getRuleMapping();
        int offsetFromBottom = ruleFileConfig.getOffsetFromBottom();


        List<RESTBatchOperation> ret = new ArrayList<RESTBatchOperation>();

        final String reqGroupName = ruleOp.getGroupName();
        if(! availableGroups.containsKey(reqGroupName.toUpperCase())) {
            LOGGER.warn("Adding new group '"+reqGroupName+"'" );
            availableGroups.put(reqGroupName.toUpperCase(), reqGroupName);
            RESTBatchOperation op = RESTBatchOperationFactory.createGroupInputOp(reqGroupName);
            ret.add(op);
        }
        final String realGroupName = availableGroups.get(ruleOp.getGroupName().toUpperCase());

        List<ServiceRequest> mapping = ruleMapping.get(ruleOp.getVerb());
        if(mapping == null) {
            LOGGER.error("Unknown verb in " + ruleOp);
            throw new IllegalArgumentException("Unknown verb in " + ruleOp);

        } else {
            for (ServiceRequest serviceRequest : mapping) {
            // running the list in reverse order since we'll add the rules using offsetFromBottom,
            // so they will result in a reversed order again
//            for (int i = mapping.size() -1; i>=0; i--) {
//                ServiceRequest serviceRequest = mapping.get(i);
                RESTBatchOperation restOp = new RESTBatchOperation();
                restOp.setService(RESTBatchOperation.ServiceName.rules);
                restOp.setType(RESTBatchOperation.TypeName.insert);

                RESTInputRule rule = new RESTInputRule();
                rule.setGroupName(realGroupName);
                rule.setLayer(ruleOp.getLayerName());
                rule.setService(serviceRequest.getService());
                rule.setRequest(serviceRequest.getRequest());
                if(serviceRequest.getGrant()==Type.allow)
                    rule.setGrant(GrantType.ALLOW);
                else if(serviceRequest.getGrant()==Type.deny)
                    rule.setGrant(GrantType.DENY);
                else {
                    throw new IllegalArgumentException("Unexpected grant type in " + serviceRequest + " for " + ruleOp);
                }
                rule.setPosition(new RESTInputRule.RESTRulePosition(RESTInputRule.RESTRulePosition.RulePosition.offsetFromBottom, offsetFromBottom));

                restOp.setPayload(rule);
                ret.add(restOp);
            }        
        }

        return ret;
    }

}
